import numpy as np

import logging

import os
from pathlib import Path

import pytest

import rasterio

from fieldfinder.spectral_index import SpectralIndex

logger = logging.getLogger(__name__)

TEST_INPUT_GEOTIFF = "test_AnalyticMS_8b.tif"

TEST_OUTPUT_GEOTIFF = "test_mask_ndvi_0_65.tif"

_here = Path(os.path.abspath(os.path.dirname(__file__)))
_test_data_path = _here / "data"

red_band_test = np.array(
    [
        [2994, 3021, 3014, 2996, 2975],
        [2970, 2972, 2975, 2984, 3022],
        [3006, 2992, 2965, 2960, 3005],
        [3019, 2974, 2927, 2905, 2911],
        [2976, 2922, 2861, 2844, 2864],
    ],
    dtype=np.uint16,
)

nir_band_test = np.array(
    [
        [6443, 6496, 6574, 6665, 6735],
        [6532, 6582, 6572, 6589, 6614],
        [6626, 6632, 6591, 6553, 6523],
        [6639, 6676, 6639, 6572, 6515],
        [6546, 6634, 6655, 6641, 6610],
    ],
    dtype=np.uint16,
)

ndvi_values_test = np.array(
    [
        [0.36547632, 0.36513607, 0.37129746, 0.37977435, 0.38722966],
        [0.37486845, 0.37785221, 0.37676757, 0.37657996, 0.37276878],
        [0.37583056, 0.37822111, 0.37944747, 0.37769368, 0.36922754],
        [0.3748188, 0.38362694, 0.38804098, 0.38693679, 0.3823467],
        [0.37492124, 0.38844705, 0.39869693, 0.40031629, 0.39539793],
    ]
)

ndvi_mask_0_65_test = np.array(
    [
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
    ]
)

ndvi_mask_0_38_test = np.array(
    [
        [0, 0, 0, 0, 255],
        [0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0],
        [0, 255, 255, 255, 255],
        [0, 255, 255, 255, 255],
    ]
)

ndvi_mask_0_65_reproject = np.array(
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
    ],
    dtype=np.uint8,
)


def test_file_constructor_type():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)
    assert isinstance(index, SpectralIndex)


def test_file_constructor_values():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)
    np.testing.assert_array_almost_equal(index.values, ndvi_values_test)


def test_file_constructor_invalid_type():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF

    with pytest.raises(Exception):
        SpectralIndex(img_path, index_type="invalid")


def test_file_constructor_invalid_geotiff():

    # use the one-band output test image to test
    img_path = _test_data_path / TEST_OUTPUT_GEOTIFF

    with pytest.raises(Exception):
        SpectralIndex(img_path)


def test_mask_065():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)
    mask = index.get_mask(threshold=0.65)
    np.testing.assert_array_equal(mask, ndvi_mask_0_65_test)


def test_mask_036():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)
    mask = index.get_mask(threshold=0.38)
    np.testing.assert_array_equal(mask, ndvi_mask_0_38_test)


def test_write_mask():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)

    output_test_file = "test_output.tif"
    index.write_mask(output_test_file, threshold=0.65, out_proj="EPSG:4326")

    with rasterio.open(output_test_file) as src:

        output_mask_test = src.read(1)

    np.testing.assert_array_equal(ndvi_mask_0_65_reproject, output_mask_test)


def test_output_projection():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    index = SpectralIndex(img_path)

    output_test_file = "test_output.tif"
    index.write_mask(output_test_file, threshold=0.65, out_proj="EPSG:4326")

    with rasterio.open(output_test_file) as src:
        src_meta = src.meta.copy()

    assert src_meta["crs"] == rasterio.crs.CRS.from_epsg(4326)


def test_create_mask_file():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    output_test_file = "test_output.tif"
    out_proj = "EPSG:4326"

    SpectralIndex.create_mask_file(
        img_path, output_test_file, threshold=0.65, out_proj=out_proj
    )

    with rasterio.open(output_test_file) as src:

        output_mask_test = src.read(1)

    np.testing.assert_array_equal(ndvi_mask_0_65_reproject, output_mask_test)


def test_create_mask_file_crs():
    img_path = _test_data_path / TEST_INPUT_GEOTIFF
    output_test_file = "test_output.tif"
    out_proj = "EPSG:4326"

    SpectralIndex.create_mask_file(
        img_path, output_test_file, threshold=0.65, out_proj=out_proj
    )

    with rasterio.open(output_test_file) as src:

        src_meta = src.meta.copy()

    assert src_meta["crs"] == rasterio.crs.CRS.from_epsg(4326)



# -*- coding: utf-8 -*-
"""
***************************************************************************
    spectral_index.py
    ---------------------
    Date                 : December 2022
    Author                : Brendon Hall

Model for creating and managing spectral indices.
***************************************************************************
"""
__author__ = "Brendon Hall"
__date__ = "December 2022"


from typing import Tuple
import numpy as np

import warnings


import rasterio
from rasterio.warp import calculate_default_transform, reproject, Resampling

from .constants import SPECTRAL_INDICES, ANALYTICMS_8B_INDEX_MAP
from .utils import is_valid_geotiff, calculate_ndvi


class SpectralIndex:
    """This is the main model class for creating a spectral index from an
    8-band PlanetScope OrthoScene.

    SpectralIndex objects can be created in two ways: with just a filename and
    with a filename and an xml metadata file.

    Attributes
    ----------
    index_type : str
        string indicating the name of the spectral index (NDVI is default)
    values : np.ndarray (2d)
        2d array that holds the values of the calculated spectral index.
    src_meta : dict
        Meta-data from the source GeoTiff file.
    src_bounds : tuple
        Spatial bounds of the source GeoTiff file.
    src_filename: str
        Filename of the source GeoTiff image

    Methods
    -------
    get_mask() - get the mask based on the index values and a given
                 threshold
    write_mask() - write a mask based the index values and a threshold
                   to a file, reprojecting if needed.

    Public Static methods
    ---------------------
    create_mask_file() - helper function to create a mask file based on an
                         index from a geotiff.  Output file will be reprojected
                         if necessary.
    """

    def __init__(self, src_filename, index_type="ndvi") -> None:
        """
        Create a spectral index based on a source GeoTiff file.  Note that this
        constructor will use the raw DNs in the GeoTiff file.  For
        PlanetScope OrthoScene images, these will be radiance values.  This
        may be appropriate for an analysis of a single image. If analysis
        across a series of images is to be performed, these values should be
        converted to TOA reflectance using coefficients in the metadata that
        normally accompanies Planetscope GeoTiff files.

        :param src_filename: filename of the source image.
        :type dst_filename: str
        :param index_type: spectral index name to calculate, defaults to 'NDVI'
        :type index_type: str, optional
        """
        if index_type.upper() in SPECTRAL_INDICES:
            self.index_type = index_type.upper()
            print(f"Creating {self.index_type} spectral index...")
        else:
            raise ValueError(
                f"Spectral index type must be one of {SPECTRAL_INDICES}"
            )

        # For now, just ensure source image has 8 bands
        if not is_valid_geotiff(src_filename):
            raise ValueError(
                f"{src_filename} is not a valid input file for this analysis."
            )
        else:
            self.src_filename = src_filename

        self.values, self.src_meta, self.src_bounds = self.calculate_index()

    def get_mask(self, threshold: float = 0) -> np.ndarray:
        """
        Generate a binary mask based on the spectral index values.  Mask
        pixels will be 255 if the spectral index values are greater than the
        threshold parameter, and 0 otherwise.

        :param threshold: index threshold values, defaults to 0
        :type threshold: float, optional
        :return: binary mask of the spectral index.
        :rtype: np.ndarry
        """

        min_value = np.nanmin(self.values)
        max_value = np.nanmax(self.values)

        if threshold < min_value or threshold > max_value:
            warnings.warn(
                f"Threshold value of {threshold} is outside the range of the"
                " index values [{min_value} to {max_value}]."
            )

        masked_index = np.where(self.values >= threshold, 255, 0)
        masked_index = masked_index.astype(np.uint8)

        return masked_index

    def write_mask(
        self, dst_filename: str, threshold: float = None, out_proj: str = None
    ) -> None:
        """
        Generate an image mask of the index based on threshold.  Reproject if
        necessary.

        :param dst_filename: filename of the destination image.
        :type dst_filename: str
        :param threshold: index value cutoff, defaults to None
        :type threshold: float, optional
        :param out_proj: CRS to reproject to.  If None, will output to the same
        CRS as the input image. Defaults to None
        :type out_proj: str, optional
        """

        # handle the case were there is no out_proj (don't need to reproject)

        dst_meta = self.src_meta.copy()

        masked_index = self.get_mask(threshold)

        if out_proj:
            transform, width, height = calculate_default_transform(
                self.src_meta["crs"],
                out_proj,
                self.src_meta["width"],
                self.src_meta["height"],
                *self.src_bounds,
            )

            dst_meta.update(
                {
                    "crs": out_proj,
                    "transform": transform,
                    "width": width,
                    "height": height,
                    "count": 1,
                    "dtype": "uint8",
                    "nodata": 0,
                }
            )

            with rasterio.open(dst_filename, "w", **dst_meta) as dst:

                reproject(
                    source=masked_index,
                    destination=rasterio.band(dst, 1),
                    src_transform=self.src_meta["transform"],
                    src_crs=self.src_meta["crs"],
                    dst_transform=transform,
                    out_proj=out_proj,
                    resampling=Resampling.nearest,
                )
        else:
            dst_meta.update(
                {
                    "count": 1,
                    "dtype": "uint8",
                    "nodata": 0,
                }
            )
            with rasterio.open(dst_filename, "w", **dst_meta) as dst:
                dst.write(masked_index, 1)

    def calculate_index(self) -> Tuple[np.ndarray, dict, tuple]:
        """Private method to read src_geotiff file, extract meta data and
           compute the relevant spectral index.

        :return: Tuple with index array, source meta data and source bounds
        :rtype: Tuple(np.ndarray, dict, tuple)
        """

        with rasterio.open(self.src_filename) as src:
            src_meta = src.meta.copy()
            src_bounds = src.bounds

            if self.index_type == "NDVI":
                red = src.read(ANALYTICMS_8B_INDEX_MAP["Red"])
                nir = src.read(ANALYTICMS_8B_INDEX_MAP["NIR"])

                ndvi = calculate_ndvi(red=red, nir=nir)

                return ndvi, src_meta, src_bounds

    @staticmethod
    def create_mask_file(
        input_file: str,
        output_file: str,
        threshold: float = 0,
        out_proj: str = None,
        index_type: str = "ndvi",
    ):
        """Helper function to write a spectral index mask file and reproject
        in one step.

        :param input_file: Source 8-band PlanetScope OrthoScene Geotiff.
        :type input_file: str
        :param output_file: Destination geotiff file to store the mask file.
        :type output_file: str
        :param threshold: Threshold value for identifying index mask cutoff,
                          defaults to 0
        :type threshold: float, optional
        :param out_proj: Output projection CRS, in EPSG code format,
                         defaults to None if no reprojection necessary
        :type out_proj: str, optional
        :param index_type: Spectral Index to calculate, defaults to "ndvi"
        :type index_type: str, optional
        """
        index = SpectralIndex(input_file, index_type)
        index.write_mask(output_file, threshold=threshold, out_proj=out_proj)



"""
***************************************************************************
    utils.py
    ---------------------
    Date                 : December 2022
    Author                : Brendon Hall

Utility and helper functions for field finder module.
***************************************************************************
"""
__author__ = "Brendon Hall"
__date__ = "December 2022"

import numpy as np

import rasterio

from .constants import ANALYTICMS_8B_INDEX_MAP


def is_valid_geotiff(src_filename: str) -> bool:
    """
    Validation method to ensure src_geotiff is a valid 8 band AnalyticMS image.
    Keep this simple for now - just check to see if the image has the
    appropriate number of bands.

    :param src_filename: filename of the src geotiff
    :type src_filename: str
    """
    with rasterio.open(src_filename) as src:
        num_bands = src.meta["count"]
        if num_bands != len(ANALYTICMS_8B_INDEX_MAP):
            raise ValueError(
                f"{src_filename} is not a valid 8-band source image."
            )

    return True


def calculate_ndvi(red: np.ndarray, nir: np.ndarray) -> np.ndarray:
    """Calculate ndvi accoring to the formula
        ndvi = (nir-red) / (nir + red)

    See https://en.wikipedia.org/wiki/Normalized_difference_vegetation_index
    for reference.

    :param red: 2d numpy array of red image values
    :type red: np.ndarray
    :param nir: 2d numpy array of nir image values
    :type nir: np.ndarray
    :return: 2d array of calculated ndvi values
    :rtype: np.ndarray
    """

    # Allow division by zero
    np.seterr(divide="ignore", invalid="ignore")

    ndvi = (nir.astype(float) - red.astype(float)) / (nir + red)

    return ndvi
